#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <memory>
#include <thread>

#include "OptionParser.h"
#include "Definitions.h"
#include "util/Timer.h"
#include "ltl/LTLParser.h"
#include "ltl/Specification.h"
#include "mealy/MealyMachine.h"
#include "pg/PGArena.h"
#include "pg/PGSolver.h"
#include "pg/PGSISolver.h"
#include "aig/AigerConstructor.h"

#include "owl.h"

std::string construct_classpath(const strix::Options& options) {
    const boost::filesystem::path path(options.owl_jar_dir);
    std::stringstream cp;
    cp << "-Djava.class.path=\":";

    if (boost::filesystem::exists(path) && boost::filesystem::is_directory(path)) {
        boost::filesystem::directory_iterator it(path);
        boost::filesystem::directory_iterator endit;
        while (it != endit) {
            if (boost::filesystem::is_regular_file(*it) && it->path().extension() == ".jar") {
                cp << it->path().string() << ":";
            }
            it++;
        }
    }

    cp << "\"";
    return cp.str();
}


std::vector<mealy::MealyMachine> check_realizability(strix::Options& options, std::ostream& out) {
    std::vector<mealy::MealyMachine> mealy_machines;

    Timer realizability_timer(options.timing);
    Timer timer(options.timing);

    timer.start("initializing OWL-JNI");
    const std::string classpath_option = construct_classpath(options);
    owl::OwlJavaVM owlJavaVM(classpath_option.c_str(), options.validate_jni, options.initial_heap_size_gb, options.max_heap_size_gb, options.aggressive_heap_optimizations);
    owl::OwlThread owl = owlJavaVM.attachCurrentThread();
    timer.stop();

    timer.start("parsing input files");
    std::vector<ltl::Specification> specifications;
    std::unique_ptr<ltl::Parser> parser(new ltl::LTLParser(owl, options.inputs, options.outputs));
    if (options.formula.empty() && options.input_files.empty()) {
        throw std::invalid_argument("No input files or formula was given.");
    }
    if (!options.formula.empty()) {
        specifications.push_back(parser->parse_string(options.formula, options.verbosity));
    }
    for (const auto& input_file : options.input_files) {
        specifications.push_back(parser->parse_string(parser->parse_file(input_file), options.verbosity));
    }
    parser.reset();
    timer.stop();

    for (const ltl::Specification& spec : specifications) {
        realizability_timer.start("checking realizability");

        if (options.verbosity >= 1) {
            std::cout << "Constructing automaton for formula ";
            spec.formula.print();
        }
        timer.start("constructing automaton tree");
        const int firstOutputVariable = spec.inputs.size();
        owl::DecomposedDPA automaton = owl.createAutomaton(spec.formula, options.simplify_formula, options.monolithic, firstOutputVariable);
        aut::AutomatonTreeStructure structure(std::move(automaton));
        pg::PGArena arena(spec.inputs.size(), spec.outputs.size(), structure, options.exploration, options.clear_queue);
        timer.stop();

        if (options.onthefly) {
            timer.start("solving game in parallel with construction of arena");
        }
        else {
            timer.start("constructing arena");
            arena.constructArena(false, options.realizability, options.verbosity);
            timer.stop();
            timer.start("solving game");
        }

        pg::PGSISolver solver(arena, options.onthefly, options.threads, options.compact_colors, options.verbosity);
        if (options.onthefly) {
            std::thread solver_thread = std::thread(&pg::PGSolver::solve, &solver);
            arena.constructArena(true, options.realizability, options.verbosity);
            solver_thread.join();
        }
        else {
            solver.solve();
        }
        timer.stop();

        Player winner = solver.getWinner();
        switch (winner) {
            case SYS_PLAYER:
                std::cout << "REALIZABLE" << std::endl;
                break;
            case ENV_PLAYER:
                std::cout << "UNREALIZABLE" << std::endl;
                break;
            case UNKNOWN:
                std::cout << "UNKNOWN" << std::endl;
                break;
        }
        realizability_timer.stop();

        if (options.verbosity >= 1) {
            arena.print_basic_info();
        }
        if (options.parity_game) {
            timer.start("writing parity game");
            arena.print(out, winner);
            timer.stop();
        }

        if (winner == SYS_PLAYER && !options.realizability) {
            timer.start("constructing Mealy machine");
            mealy::MealyMachine m(std::move(spec.inputs), std::move(spec.outputs));
            bool with_labels = solver.constructMealyMachine(m, options.labels);
            if (options.labels && !with_labels) {
                options.labels = false;
                if (options.verbosity >= 1) {
                    std::cerr << "Error: cannot construct product labels" << std::endl;
                }
            }
            mealy_machines.push_back(std::move(m));
            timer.stop();
        }
    }
    return mealy_machines;
}

void construct_solution(mealy::MealyMachine& m, strix::Options& options, std::ostream& out) {
    Timer timer(options.timing);

    if (options.verbosity >= 1) {
        std::cout << "Number of states of Mealy machine: " << m.numberOfStates() << std::endl;
    }
    // in auto mode, do not try to minimize huge Mealy machines
    if (options.minimize && options.auto_aiger && m.numberOfStates() > 10000) {
        options.minimize = false;
    }
    if (options.minimize) {
        timer.start("minimize Mealy machine");
        if (options.verbosity >= 1) {
            std::cout << "Minimizing Mealy machine" << std::endl;
        }
        if (!m.minimizeMachine(options.labels, options.verbosity)) {
            options.minimize = false;
            if (options.verbosity >= 1) {
                std::cout << "Minimizing Mealy machine failed" << std::endl;
            }
        }
        else if (options.verbosity > 1) {
            std::cout << "Number of states of minimized Mealy machine: " << m.numberOfMinStates() << std::endl;
        }
        timer.stop();
    }
    if (options.mealy) {
        if (options.dot) {
            m.print_dot(out, options.labels, options.minimize);
        }
        else {
            m.print_kiss(out, options.labels, options.minimize);
        }
    }
    else {
        timer.start("constructing AIGER circuit");
        std::shared_ptr<aig::AigerConstructor> aiger;
        if (options.auto_aiger) {
            aiger = aig::AigerConstructor::findMinimalAiger(m, options.labels, options.minimize, options.compress_circuit, options.bdd);
        }
        else {
            aiger = std::shared_ptr<aig::AigerConstructor>(new aig::AigerConstructor(m, options.labels, options.minimize));
            aiger->construct(options.compress_circuit, options.bdd);
        }
        if (options.bdd) {
            aiger->printBDDs(out);
        }
        else if (options.dot) {
            aiger->print_dot(out);
        }
        else {
            aiger->print_aiger(out, options.binary);
        }
        timer.stop();
    }
}

void synthesis(strix::Options& options) {
    Timer timer(options.timing);
    timer.start("running Strix");

    std::streambuf* buf;
    std::ofstream of;
    if (options.output_file.empty()) {
        buf = std::cout.rdbuf();
    }
    else {
        of.open(options.output_file, std::ios::out);
        if (!of) {
            throw std::invalid_argument("Could not open output file: " + options.output_file);
        }
        buf = of.rdbuf();
    }
    std::ostream out(buf);

    std::vector<mealy::MealyMachine> mealy_machines = check_realizability(options, out);
    timer.stop();

    for (mealy::MealyMachine& m : mealy_machines) {
        timer.start("constructing solution");
        construct_solution(m, options, out);
        timer.stop();
    }

    of.close();
}

int main(const int argc, const char* argv[]) {
    try {
        // Parse options
        strix::OptionParser option_parser(argc, argv);
        strix::Options options = option_parser.parse_options();

        if (options.help) {
            option_parser.print_help();
            return EXIT_SUCCESS;
        }
        else {
            synthesis(options);
        }

    }
    catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::string& e) {
        std::cerr << "Error: " << e << std::endl;
        return EXIT_FAILURE;
    }
    catch (...) {
        std::cerr << "Error: Unknown error." << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
